with
  text_io,
  ada.integer_text_io,
  ada.long_float_text_io,
  ada.numerics,
  ada.numerics.long_elementary_functions,
  terms;

procedure dodecahedron2 is

  type elem1_rec is
    record
      glob : terms.imat1 (1 .. 2);
    end record;

  type elem1_vec is array (integer range <>) of elem1_rec;

  type elem1_ptr is access elem1_vec;

  type elem2_rec is
    record
      glob : terms.imat1 (1 .. 3);
    end record;

  type elem2_vec is array (integer range <>) of elem2_rec;

  type elem2_ptr is access elem2_vec;

  ni,me,mf,nn,ne1,ne2,state : integer;
  radius,r                  : long_float;
  area,young,value          : long_float;
  stiffness,poisson         : long_float;
  edge                      : array (integer range 1 .. 30, integer range 1 .. 2) of integer;
  face,fedge                : array (integer range 1 .. 12, integer range 0 .. 4) of integer;
  xo,yo,zo                  : terms.fmat1_ptr;
  elem1                     : elem1_ptr;
  elem2                     : elem2_ptr;

  procedure scan (io : in text_io.file_type) is

    ascii : character;

  begin
    loop
      text_io.get(io,ascii);
      text_io.skip_line(io);
      exit when ascii = ':';
    end loop;
  end scan;

  function fnode (f,i,j : in integer) return integer is

    k,e : integer;

  begin
    if i = 0 and j = 0 then
      return 21 + 30 * me + (f - 1) * mf;
    end if;

    if i = ni then
      k := j / ni;
      if (j rem ni) = 0 then
        return face(f,k);
      else
        e := fedge(f,k);
        if e > 0 then
          return (e - 1) * (ni - 1) + (j - k * i) + 20;
        else
          return e * (1 - ni) - (j - k * i) + 21;
        end if;
      end if;
    else
      return (i - 1) * i / 2 * 5 + j + 22 + 30 * me + (f - 1) * mf;
    end if;
  end fnode;

  procedure facenodes is

    use
      ada.numerics,
      ada.numerics.long_elementary_functions;

    glob,k1,k2             : integer;
    norm,x0,y0,z0,ri,alpha : long_float;
    o,u,v,p1,p2            : array (integer range 1 .. 3) of long_float;
    cost,sint              : array (integer range 0 .. 4) of long_float;

  begin
    -- cost, sint
    alpha := 0.4 * pi;
    for k in 0 .. 4 loop
      cost(k) := cos(alpha * long_float(k));
      sint(k) := sin(alpha * long_float(k));
    end loop;

    for f in 1 .. 12 loop
      -- face coordinate system
      o(1) := 0.0;
      o(2) := 0.0;
      o(3) := 0.0;
      for j in 0 .. 4 loop
          glob := face(f,j);
          o(1) := o(1) + xo(glob);
          o(2) := o(2) + yo(glob);
          o(3) := o(3) + zo(glob);
      end loop;
      o(1) := 0.2 * o(1);
      o(2) := 0.2 * o(2);
      o(3) := 0.2 * o(3);

      glob := face(f,0);
      x0 := xo(glob);
      y0 := yo(glob);
      z0 := zo(glob);

      u(1) := x0 - o(1);
      u(2) := y0 - o(2);
      u(3) := z0 - o(3);
      norm := sqrt(u(1) * u(1) + u(2) * u(2) + u(3) * u(3));
      u(1) := u(1) / norm;
      u(2) := u(2) / norm;
      u(3) := u(3) / norm;

      v(1) := o(2) * z0 - o(3) * y0;
      v(2) := o(3) * x0 - o(1) * z0;
      v(3) := o(1) * y0 - o(2) * x0;
      norm := sqrt(v(1) * v(1) + v(2) * v(2) + v(3) * v(3));
      v(1) := v(1) / norm;
      v(2) := v(2) / norm;
      v(3) := v(3) / norm;

      -- face nodes
      glob := fnode(f,0,0);
      xo(glob) := o(1);
      yo(glob) := o(2);
      zo(glob) := o(3);

      for i in 1 .. ni - 1 loop
        ri := r * long_float(i) / long_float(ni);

        for j in 0 .. 5 * i - 1 loop
          k1 := j / i;
          k2 := (k1 + 1) rem 5;

          p1(1) := o(1) + ri * (cost(k1) * u(1) + sint(k1) * v(1));
          p1(2) := o(2) + ri * (cost(k1) * u(2) + sint(k1) * v(2));
          p1(3) := o(3) + ri * (cost(k1) * u(3) + sint(k1) * v(3));

          p2(1) := o(1) + ri * (cost(k2) * u(1) + sint(k2) * v(1));
          p2(2) := o(2) + ri * (cost(k2) * u(2) + sint(k2) * v(2));
          p2(3) := o(3) + ri * (cost(k2) * u(3) + sint(k2) * v(3));

          alpha := long_float(j - k1 * i) / long_float(i);
          glob := fnode(f,i,j);
          xo(glob) := (1.0 - alpha) * p1(1) + alpha * p2(1);
          yo(glob) := (1.0 - alpha) * p1(2) + alpha * p2(2);
          zo(glob) := (1.0 - alpha) * p1(3) + alpha * p2(3);
        end loop;
      end loop;
    end loop;
  end facenodes;

  procedure initia is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io,
      ada.numerics,
      ada.numerics.long_elementary_functions;

    glob,glob1,glob2,k,e : integer;
    c1,c2,c3,alpha       : long_float;
    io                   : file_type;

  begin
    open(io, in_file, "dodecahedron2.txt");

    -- ni, radius
    scan(io);
    get(io,ni);
    get(io,radius);
    skip_line(io);

    -- area, young, state, value
    scan(io);
    get(io,area);
    get(io,young);
    get(io,state);
    get(io,value);
    skip_line(io);

    -- stiffness, poisson
    scan(io);
    get(io,stiffness);
    get(io,poisson);
    skip_line(io);

    close(io);

    -- number of nodes from each edge divisions
    me := ni - 1;

    -- number of nodes from each face
    mf := (ni - 1) * ni / 2 * 5 + 1;

    -- number of nodes
    nn := 30 * ni * ni + 2;

    -- number of line elements
    ne1 := 10;

    -- number of triangle elements
    ne2 := 60 * ni * ni;

    -- radius of the polygonal face
    r := 2.0 / sqrt(3.0) / (1.0 + sqrt(5.0)) / sin(0.2 * pi) * radius;

    -- xo, yo, zo
    xo := new terms.fmat1 (1 .. nn);
    yo := new terms.fmat1 (1 .. nn);
    zo := new terms.fmat1 (1 .. nn);

    -- vertices coordinates
    c1 := (1.0 + sqrt(5.0)) / 2.0 / sqrt(3.0) * radius;
    c2 := 2.0 / (1.0 + sqrt(5.0)) / sqrt(3.0) * radius;
    c3 := radius / sqrt(3.0);

    xo(1) := 0.0;
    yo(1) := c2;
    zo(1) := c1;

    xo(2) := 0.0;
    yo(2) := -c2;
    zo(2) := c1;

    xo(3) := c3;
    yo(3) := c3;
    zo(3) := c3;

    xo(4) := -c3;
    yo(4) := c3;
    zo(4) := c3;

    xo(5) := -c3;
    yo(5) := -c3;
    zo(5) := c3;

    xo(6) := c3;
    yo(6) := -c3;
    zo(6) := c3;

    xo(7) := c1;
    yo(7) := 0.0;
    zo(7) := c2;

    xo(8) := -c1;
    yo(8) := 0.0;
    zo(8) := c2;

    xo(9) := c2;
    yo(9) := c1;
    zo(9) := 0.0;

    xo(10) := -c2;
    yo(10) := c1;
    zo(10) := 0.0;

    xo(11) := -c2;
    yo(11) := -c1;
    zo(11) := 0.0;

    xo(12) := c2;
    yo(12) := -c1;
    zo(12) := 0.0;

    xo(13) := c1;
    yo(13) := 0.0;
    zo(13) := -c2;

    xo(14) := -c1;
    yo(14) := 0.0;
    zo(14) := -c2;

    xo(15) := c3;
    yo(15) := c3;
    zo(15) := -c3;

    xo(16) := -c3;
    yo(16) := c3;
    zo(16) := -c3;

    xo(17) := -c3;
    yo(17) := -c3;
    zo(17) := -c3;

    xo(18) := c3;
    yo(18) := -c3;
    zo(18) := -c3;

    xo(19) := 0.0;
    yo(19) := c2;
    zo(19) := -c1;

    xo(20) := 0.0;
    yo(20) := -c2;
    zo(20) := -c1;

    -- edges
    edge(1,1) := 1;
    edge(1,2) := 2;

    edge(2,1) := 1;
    edge(2,2) := 3;

    edge(3,1) := 1;
    edge(3,2) := 4;

    edge(4,1) := 2;
    edge(4,2) := 5;

    edge(5,1) := 2;
    edge(5,2) := 6;

    edge(6,1) := 3;
    edge(6,2) := 7;

    edge(7,1) := 3;
    edge(7,2) := 9;

    edge(8,1) := 4;
    edge(8,2) := 8;

    edge(9,1) := 4;
    edge(9,2) := 10;

    edge(10,1) := 5;
    edge(10,2) := 8;

    edge(11,1) := 5;
    edge(11,2) := 11;

    edge(12,1) := 6;
    edge(12,2) := 7;

    edge(13,1) := 6;
    edge(13,2) := 12;

    edge(14,1) := 7;
    edge(14,2) := 13;

    edge(15,1) := 8;
    edge(15,2) := 14;

    edge(16,1) := 9;
    edge(16,2) := 10;

    edge(17,1) := 9;
    edge(17,2) := 15;

    edge(18,1) := 10;
    edge(18,2) := 16;

    edge(19,1) := 11;
    edge(19,2) := 12;

    edge(20,1) := 11;
    edge(20,2) := 17;

    edge(21,1) := 12;
    edge(21,2) := 18;

    edge(22,1) := 13;
    edge(22,2) := 15;

    edge(23,1) := 13;
    edge(23,2) := 18;

    edge(24,1) := 14;
    edge(24,2) := 16;

    edge(25,1) := 14;
    edge(25,2) := 17;

    edge(26,1) := 15;
    edge(26,2) := 19;

    edge(27,1) := 16;
    edge(27,2) := 19;

    edge(28,1) := 17;
    edge(28,2) := 20;

    edge(29,1) := 18;
    edge(29,2) := 20;

    edge(30,1) := 19;
    edge(30,2) := 20;

    -- faces
    face(1,0) := 1;
    face(1,1) := 2;
    face(1,2) := 6;
    face(1,3) := 7;
    face(1,4) := 3;

    face(2,0) := 1;
    face(2,1) := 3;
    face(2,2) := 9;
    face(2,3) := 10;
    face(2,4) := 4;

    face(3,0) := 1;
    face(3,1) := 4;
    face(3,2) := 8;
    face(3,3) := 5;
    face(3,4) := 2;

    face(4,0) := 2;
    face(4,1) := 5;
    face(4,2) := 11;
    face(4,3) := 12;
    face(4,4) := 6;

    face(5,0) := 3;
    face(5,1) := 7;
    face(5,2) := 13;
    face(5,3) := 15;
    face(5,4) := 9;

    face(6,0) := 4;
    face(6,1) := 10;
    face(6,2) := 16;
    face(6,3) := 14;
    face(6,4) := 8;

    face(7,0) := 5;
    face(7,1) := 8;
    face(7,2) := 14;
    face(7,3) := 17;
    face(7,4) := 11;

    face(8,0) := 6;
    face(8,1) := 12;
    face(8,2) := 18;
    face(8,3) := 13;
    face(8,4) := 7;

    face(9,0) := 9;
    face(9,1) := 15;
    face(9,2) := 19;
    face(9,3) := 16;
    face(9,4) := 10;

    face(10,0) := 11;
    face(10,1) := 17;
    face(10,2) := 20;
    face(10,3) := 18;
    face(10,4) := 12;

    face(11,0) := 13;
    face(11,1) := 18;
    face(11,2) := 20;
    face(11,3) := 19;
    face(11,4) := 15;

    face(12,0) := 14;
    face(12,1) := 16;
    face(12,2) := 19;
    face(12,3) := 20;
    face(12,4) := 17;

    -- elements 1
    elem1 := new elem1_vec (1 .. ne1);

    elem1(1).glob(1) := 2;
    elem1(1).glob(2) := 19;

    elem1(2).glob(1) := 5;
    elem1(2).glob(2) := 16;

    elem1(3).glob(1) := 6;
    elem1(3).glob(2) := 15;

    elem1(4).glob(1) := 7;
    elem1(4).glob(2) := 8;

    elem1(5).glob(1) := 13;
    elem1(5).glob(2) := 14;

    elem1(6).glob(1) := 1;
    elem1(6).glob(2) := 12;

    elem1(7).glob(1) := 3;
    elem1(7).glob(2) := 18;

    elem1(8).glob(1) := 4;
    elem1(8).glob(2) := 11;

    elem1(9).glob(1) := 9;
    elem1(9).glob(2) := 20;

    elem1(10).glob(1) := 10;
    elem1(10).glob(2) := 17;

    -- coordinates of points along edges
    for e in 1 .. 30 loop
      if edge(e,2) > edge(e,1) then
        glob1 := edge(e,1);
        glob2 := edge(e,2);
      else
        glob1 := edge(e,2);
        glob2 := edge(e,1);
      end if;

      for j in 1 .. ni - 1 loop
        alpha := long_float(j) / long_float(ni);
        glob := (e - 1) * (ni - 1) + j + 20;
        xo(glob) := (1.0 - alpha) * xo(glob1) + alpha * xo(glob2);
        yo(glob) := (1.0 - alpha) * yo(glob1) + alpha * yo(glob2);
        zo(glob) := (1.0 - alpha) * zo(glob1) + alpha * zo(glob2);
      end loop;
    end loop;

    -- edge numbers of a face
    for f in 1 .. 12 loop
      for k in 0 .. 4 loop
        glob1 := face(f,k);
        glob2 := face(f,(k + 1) rem 5);

        e := 0;
        loop
          e := e + 1;

          if glob1 = edge(e,1) and glob2 = edge(e,2) then
            fedge(f,k) := e;
            exit;
          end if;

          if glob2 = edge(e,1) and glob1 = edge(e,2) then
            fedge(f,k) := -e;
            exit;
          end if;
        end loop;
      end loop;
    end loop;

    -- face node coordinates
    facenodes;

    -- elements 2
    elem2 := new elem2_vec (1 .. ne2);

    for f in 1 .. 12 loop
      e := 5 * ni * ni * (f - 1);

      for j in 0 .. 3 loop
        e := e + 1;
        elem2(e).glob(1) := fnode(f,0,0);
        elem2(e).glob(2) := fnode(f,1,j);
        elem2(e).glob(3) := fnode(f,1,j + 1);
      end loop;

      e := e + 1;
      elem2(e).glob(1) := fnode(f,0,0);
      elem2(e).glob(2) := fnode(f,1,4);
      elem2(e).glob(3) := fnode(f,1,0);

      for i in 2 .. ni loop
        for j in 0 .. 5 * i - 2 loop
          k := j / i;
          e := e + 1;
          elem2(e).glob(1) := fnode(f,i - 1,j - k);
          elem2(e).glob(2) := fnode(f,i,j);
          elem2(e).glob(3) := fnode(f,i,j + 1);
        end loop;

        e := e + 1;
        elem2(e).glob(1) := fnode(f,i - 1,0);
        elem2(e).glob(2) := fnode(f,i,5 * i - 1);
        elem2(e).glob(3) := fnode(f,i,0);
      end loop;

      for i in 1 .. ni - 1 loop
        for j in 0 .. 5 * i - 2 loop
          k := j / i;
          e := e + 1;
          elem2(e).glob(1) := fnode(f,i + 1,j + 1 + k);
          elem2(e).glob(2) := fnode(f,i,j + 1);
          elem2(e).glob(3) := fnode(f,i,j);
        end loop;

        e := e + 1;
        elem2(e).glob(1) := fnode(f,i + 1,5 * i + 4);
        elem2(e).glob(2) := fnode(f,i,0);
        elem2(e).glob(3) := fnode(f,i,5 * i - 1);
      end loop;
    end loop;
  end initia;

  procedure input is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    maxls,maxminor,maxlist : integer;
    lstol,mntol,mnref      : long_float;
    io                     : file_type;

  begin
    -- solution parameters
    lstol := 1.0E-01;
    maxls := 20;
    mntol := 1.0E-06;
    mnref := 1.0;
    maxminor := 30 * nn;
    maxlist := 3;

    create(io, out_file, "input.txt");

    put(io,"        lstol     maxls             mntol             mnref  maxminor   maxlist");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,lstol,2,6,3);
    put(io,"     ");
    put(io,maxls,5);
    put(io,"     ");
    put(io,mntol,2,6,3);
    put(io,"     ");
    put(io,mnref,2,6,3);
    put(io,"     ");
    put(io,maxminor,5);
    put(io,"     ");
    put(io,maxlist,5);
    new_line(io);

    -- node coordinates
    new_line(io);
    new_line(io);
    put(io," node           coord x           coord y           coord z");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,nn,5);
    new_line(io);

    new_line(io);
    for glob in 1 .. nn loop
      put(io,glob,5);
      put(io,"     ");
      put(io,xo(glob),2,6,3);
      put(io,"     ");
      put(io,yo(glob),2,6,3);
      put(io,"     ");
      put(io,zo(glob),2,6,3);
      new_line(io);
    end loop;

    -- elements 1
    new_line(io);
    new_line(io);
    put(io," elem      node      node              area             young");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,elem1'last,5);
    new_line(io);

    new_line(io);
    for e in elem1'range loop
      put(io,e,5);
      for loc in 1 .. 2 loop
        put(io,"     ");
        put(io,elem1(e).glob(loc),5);
      end loop;
      put(io,"     ");
      put(io,area,2,6,3);
      put(io,"     ");
      put(io,young,2,6,3);
      new_line(io);
    end loop;

    -- elements 2
    new_line(io);
    new_line(io);
    put(io," elem      node      node      node         stiffness           poisson");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,elem2'last,5);
    new_line(io);

    new_line(io);
    for e in elem2'range loop
      put(io,e,5);
      for loc in 1 .. 3 loop
        put(io,"     ");
        put(io,elem2(e).glob(loc),5);
      end loop;
      put(io,"     ");
      put(io,stiffness,2,6,3);
      put(io,"     ");
      put(io,poisson,2,6,3);
      new_line(io);
    end loop;

    -- known displacements
    new_line(io);
    new_line(io);
    put(io," node      axis             displ");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,"    0");
    new_line(io);

    -- set cable element state
    new_line(io);
    new_line(io);
    case state is
      when 1 =>
        put(io," elem     state               cut");
        new_line(io);

      when 2 =>
        put(io," elem     state           tension");
        new_line(io);

      when others =>
        null;
    end case;
    put(io,":");
    new_line(io);
    put(io,elem1'last,5);
    new_line(io);

    new_line(io);
    for e in elem1'range loop
      put(io,e,5);
      put(io,"     ");
      put(io,state,5);
      put(io,"     ");
      put(io,value,2,6,3);
      new_line(io);
    end loop;

    -- loading
    new_line(io);
    new_line(io);
    put(io," node      axis             force");
    new_line(io);
    put(io,":");
    new_line(io);
    put(io,"    0");
    new_line(io);

    close(io);
  end input;

  procedure autocad is

    use
      text_io,
      ada.integer_text_io,
      ada.long_float_text_io;

    glob : integer;
    h    : long_float;
    io   : file_type;

  begin
    -- text height
    h := 0.075 * radius;

    create(io, out_file, "dodecahedron2.scr");

    -- set running object snap mode to NONE
    put(io,"osmode 0");
    new_line(io);

    -- dodecahedron
    put(io,"layer make Dodecahedron ");
    new_line(io);

    for e in 1 .. 30 loop
      put(io,"3dpoly ");
      for loc in 1 .. 2 loop
        glob := edge(e,loc);
        put(io,xo(glob),1,8,3);
        put(io,",");
        put(io,yo(glob),1,8,3);
        put(io,",");
        put(io,zo(glob),1,8,3);
        put(io," ");
      end loop;
      new_line(io);
    end loop;

    -- dodecahedron
    put(io,"layer make Vertices ");
    new_line(io);

    for glob in 1 .. 20 loop
      put(io,"ucs za ");
      put(io,xo(glob),1,8,3);
      put(io,",");
      put(io,yo(glob),1,8,3);
      put(io,",");
      put(io,zo(glob),1,8,3);
      put(io," ");
      put(io,2.0 * xo(glob),1,8,3);
      put(io,",");
      put(io,2.0 * yo(glob),1,8,3);
      put(io,",");
      put(io,2.0 * zo(glob),1,8,3);
      new_line(io);

      put(io,"text j m 0.0,0.0,0.0 ");
      put(io,h,1,8,3);
      put(io," 0 ");
      put(io,glob,0);
      new_line(io);

      put(io,"ucs w");
      new_line(io);
    end loop;

    -- elements 2
    put(io,"layer make Elem2 ");
    new_line(io);

    for e in elem2'range loop
      put(io,"3dface ");
      for loc in 1 .. 3 loop
        glob := elem2(e).glob(loc);
        put(io,xo(glob),1,8,3);
        put(io,",");
        put(io,yo(glob),1,8,3);
        put(io,",");
        put(io,zo(glob),1,8,3);
        put(io," ");
      end loop;
      put(io," ");
      new_line(io);
    end loop;

    -- elements 1
    put(io,"layer make Elem1 ");
    new_line(io);

    put(io,"-color");
    new_line(io);
    put(io,"green");
    new_line(io);

    for e in elem1'range loop
      put(io,"3dpoly ");
      for loc in 1 .. 2 loop
        glob := elem1(e).glob(loc);
        put(io,xo(glob),1,8,3);
        put(io,",");
        put(io,yo(glob),1,8,3);
        put(io,",");
        put(io,zo(glob),1,8,3);
        put(io," ");
      end loop;
      new_line(io);
    end loop;

    put(io,"-color");
    new_line(io);
    put(io,"white");
    new_line(io);

    -- autocad
    put(io,"layer set 0 ");
    new_line(io);

    put(io,"layer off Dodecahedron ");
    new_line(io);

    put(io,"layer off Vertices ");
    new_line(io);

    put(io,"zoom ext ext");
    new_line(io);

    close(io);
  end autocad;

begin
  initia;
  input;
  autocad;
end dodecahedron2;
